import logging
import typing as t
from collections import Counter


class CounterHook:

    prehook_priority: int = 0
    posthook_priority: int = 0
    event_counter: t.Counter[str]

    def __init__(self):
        self.event_counter = Counter()

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> None:
        self.event_counter[event_name] += 1

    def posthook(self, event_name: str, result: t.Any, context: None) -> None:
        # Posthook does nothing
        pass

    def reset(self):
        self.event_counter.clear()

    def log_results(self, logger: logging.Logger):
        logger.info("Calls to events recorded by %s:", self.__class__.__name__)
        for (event, count) in self.event_counter.items():
            logger.info("    %s: %d", event, count)



# Tests for ProfilerHook

from seagrass import Auditor
from seagrass.hooks import ProfilerHook
import time
import unittest


class ProfilerHookTestCase(unittest.TestCase):
    def test_hook_function(self):
        # Test only works for Python >= 3.9 due to the use of StatsProfile
        try:
            from pstats import StatsProfile  # noqa: F401
        except ImportError:
            self.skipTest("Test disabled for Python < 3.9")

        auditor = Auditor()
        hook = ProfilerHook()

        # Note: could just as easily use auditor.wrap(time.sleep, ...) here
        # but the name of time.sleep is slightly mangled in the resulting
        # StatsProfile that we generate, which complicates testing.
        @auditor.decorate("test.sleep", hooks=[hook])
        def ausleep(*args):
            time.sleep(*args)

        with auditor.audit():
            for _ in range(10):
                ausleep(0.001)

        # Get profiler information for ausleep
        stats_profile = hook.get_stats().get_stats_profile()
        ausleep_profile = stats_profile.func_profiles["ausleep"]
        self.assertEqual(ausleep_profile.ncalls, "10")

        # Profiler information should be reset after hook.reset() is called
        hook.reset()
        with self.assertRaises(Exception):
            hook.get_stats()

        with auditor.audit():
            ausleep(0.01)

        stats_profile = hook.get_stats().get_stats_profile()
        ausleep_profile = stats_profile.func_profiles["ausleep"]
        self.assertEqual(ausleep_profile.ncalls, "1")
        self.assertAlmostEqual(ausleep_profile.cumtime, 0.01, delta=0.005)



import sys
from seagrass.base import ProtoHook
from typing import Any, Callable, List, Optional, Protocol


class EventFnProtocol(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        pass


class Event:
    """A wrapped function that is under audit."""

    # Use __slots__ since feasibly users may want to create a large
    # number of events
    __slots__ = [
        "func",
        "enabled",
        "name",
        "raise_audit_event",
        "hooks",
        "prehook_audit_event_name",
        "posthook_audit_event_name",
        "__prehook_execution_order",
        "__posthook_execution_order",
    ]

    func: EventFnProtocol
    enabled: bool
    name: str
    raise_audit_event: bool
    hooks: List[ProtoHook]
    prehook_audit_event_name: str
    posthook_audit_event_name: str
    __prehook_execution_order: List[int]
    __posthook_execution_order: List[int]

    def __init__(
        self,
        func: Callable,
        name: str,
        enabled: bool = True,
        hooks: List[ProtoHook] = [],
        raise_audit_event: bool = False,
        prehook_audit_event_name: Optional[str] = None,
        posthook_audit_event_name: Optional[str] = None,
    ):
        self.func = func
        self.enabled = enabled
        self.name = name
        self.raise_audit_event = raise_audit_event
        self.hooks = hooks

        if prehook_audit_event_name is None:
            prehook_audit_event_name = f"prehook:{name}"
        if posthook_audit_event_name is None:
            posthook_audit_event_name = f"posthook:{name}"

        self.prehook_audit_event_name = prehook_audit_event_name
        self.posthook_audit_event_name = posthook_audit_event_name

        # Set the order of execution for prehooks and posthooks.
        # - Prehooks are ordered by ascending priority, then ascending list position
        # - Posthooks are ordered by descending priority, then descending list position
        self.__prehook_execution_order = sorted(
            range(len(hooks)), key=lambda i: (hooks[i].prehook_priority, i)
        )
        self.__posthook_execution_order = sorted(
            range(len(hooks)), key=lambda i: (-hooks[i].posthook_priority, -i)
        )

    def __call__(self, *args, **kwargs):
        if not self.enabled:
            # We just return the result of the wrapped function
            return self.func(*args, **kwargs)

        if self.raise_audit_event:
            sys.audit(self.prehook_audit_event_name, args, kwargs)

        prehook_contexts = {}
        for hook_num in self.__prehook_execution_order:
            hook = self.hooks[hook_num]
            context = hook.prehook(self.name, args, kwargs)
            prehook_contexts[hook_num] = context

        result = self.func(*args, **kwargs)

        for hook_num in self.__posthook_execution_order:
            hook = self.hooks[hook_num]
            hook.posthook(self.name, result, prehook_contexts[hook_num])

        if self.raise_audit_event:
            sys.audit(self.posthook_audit_event_name, result)

        return result



import functools
import logging
import typing as t
from contextlib import contextmanager
from seagrass.base import LoggableHook, ProtoHook
from seagrass.events import Event

# Global variable that keeps track of the auditor's logger for the
# current auditing context.
_audit_logger_stack: t.List[logging.Logger] = []


def get_audit_logger() -> t.Optional[logging.Logger]:
    if len(_audit_logger_stack) == 0:
        return None
    else:
        return _audit_logger_stack[-1]


class Auditor:
    """
    An auditing instance that allows you to dynamically audit and profile
    code.
    """

    logger: logging.Logger
    events: t.Dict[str, Event]
    event_wrappers: t.Dict[str, t.Callable]
    hooks: t.Set[ProtoHook]
    __enabled: bool = False

    def __init__(self, logger: t.Union[str, logging.Logger] = "seagrass"):
        """Create a new Auditor instance."""
        if isinstance(logger, logging.Logger):
            self.logger = logger
        else:
            self.logger = logging.getLogger("seagrass")

        self.events = dict()
        self.event_wrappers = dict()
        self.hooks = set()

    def toggle_auditing(self, mode: bool):
        """Enable or disable auditing."""
        self.__enabled = mode

    @property
    def is_enabled(self):
        return self.__enabled

    @contextmanager
    def audit(self):
        """Create a new context within which the auditor is enabled."""
        try:
            self.toggle_auditing(True)
            _audit_logger_stack.append(self.logger)
            yield None
        finally:
            self.toggle_auditing(False)
            _audit_logger_stack.pop()

    def wrap(
        self,
        func: t.Callable,
        label: str,
        hooks: t.Optional[t.List[ProtoHook]] = None,
        **kwargs,
    ) -> t.Callable:
        """Wrap a function with a new auditing event."""

        if label in self.events:
            raise ValueError(
                f"An event with the label {label!r} has already been defined"
            )

        hooks = [] if hooks is None else hooks

        # Add hooks to the Auditor's `hooks` set
        for hook in hooks:
            self.hooks.add(hook)

        new_event = Event(func, label, hooks=hooks, **kwargs)
        self.events[label] = new_event

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not self.is_enabled:
                return new_event.func(*args, **kwargs)
            else:
                return new_event(*args, **kwargs)

        return wrapper

    def decorate(
        self,
        label: str,
        **kwargs,
    ) -> t.Callable:
        """A function decorator that tells the auditor to monitor the decorated function."""

        def wrapper(func):
            return self.wrap(func, label, **kwargs)

        return wrapper

    def toggle_event(self, label: str, enabled: bool) -> None:
        """Toggle whether or not an event is enabled."""
        self.events[label].enabled = enabled

    def log_results(self):
        """Log results stored by hooks by calling `log_results` on all LoggableHooks."""
        for hook in self.hooks:
            if isinstance(hook, LoggableHook):
                hook.log_results(self.logger)



import logging
import typing as t
from seagrass import get_audit_logger

log_input_t = str
prehook_msg_t = t.Callable[[str, t.Tuple[t.Any, ...], t.Dict[str, t.Any]], log_input_t]
posthook_msg_t = t.Callable[[str, t.Any], log_input_t]


class LoggingHook:
    """A hook that emits a new log whenever it gets called."""

    prehook_priority: int = 0
    posthook_priority: int = 0

    loglevel: int
    prehook_msg: t.Optional[prehook_msg_t]
    posthook_msg: t.Optional[posthook_msg_t]

    def __init__(
        self,
        prehook_msg: t.Optional[prehook_msg_t] = None,
        posthook_msg: t.Optional[posthook_msg_t] = None,
        loglevel: int = logging.DEBUG,
    ):
        if prehook_msg is None and posthook_msg is None:
            raise ValueError(
                "At least one of the keyword arguments prehook_msg and posthook_msg "
                + "must be specified and not equal to None"
            )

        self.prehook_msg = prehook_msg
        self.posthook_msg = posthook_msg
        self.loglevel = loglevel

    def prehook(
        self,
        event_name: str,
        args: t.Tuple[t.Any, ...],
        kwargs: t.Dict[str, t.Any],
    ) -> None:
        if self.prehook_msg is None:
            pass
        else:
            if (logger := get_audit_logger()) is not None:
                logger.log(self.loglevel, self.prehook_msg(event_name, args, kwargs))

    def posthook(
        self,
        event_name: str,
        result: t.Any,
        context: None,
    ) -> None:
        if self.posthook_msg is None:
            pass
        else:
            if (logger := get_audit_logger()) is not None:
                logger.log(self.loglevel, self.posthook_msg(event_name, result))



# Tests for Auditor creation and basic functionality

import logging
import unittest
from io import StringIO
from seagrass import Auditor


class CreateAuditorTestCase(unittest.TestCase):
    """Tests for creating a new Auditor instance."""

    def test_create_auditor_with_logger(self):
        # Create a new Auditor with a custom logger
        self.logging_output = StringIO()

        self.logger_name = "seagrass.test"
        self.logger = logging.getLogger(self.logger_name)
        self.logger.setLevel(logging.DEBUG)
        fh = logging.StreamHandler(self.logging_output)
        fh.setLevel(logging.INFO)

        formatter = logging.Formatter("%(message)s")
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)

        auditor = Auditor(logger=self.logger)
        auditor.logger.info("Hello, world!")
        auditor.logger.debug("This message shouldn't appear")

        self.logging_output.seek(0)
        lines = self.logging_output.readlines()
        self.assertEqual(lines, ["Hello, world!\n"])

    def test_define_two_events_with_the_same_name(self):
        # If we try to create two auditing events with the same name, we should
        # get an error.
        auditor = Auditor()

        @auditor.decorate("test.foo")
        def foo_1():
            return

        with self.assertRaises(ValueError):

            @auditor.decorate("test.foo")
            def foo_2():
                return


if __name__ == "__main__":
    unittest.main()



import logging
import sys
import typing as t
import warnings


class FileOpenHook:
    """An event hook for tracking calls to the Python standard
    library's `open` function."""

    # Give this hook slightly higher priority by default so that
    # we can avoid counting calls to open that occur in other
    # hooks.
    prehook_priority: int = 3
    posthook_priority: int = 3

    # Boolean variables that indicate whether we should track the
    # mode that files were opened with, and whether we should track
    # the flags those files were opened with.
    track_mode: bool
    track_flags: bool

    file_open_counter: t.Counter[str]
    __enabled: bool = False

    def __init__(self, track_mode: bool = True, track_flags: bool = False):
        self.file_open_counter = t.Counter[str]()
        self.track_mode = track_mode
        self.track_flags = track_flags

        # Add the __sys_audit_hook closure as a new audit hook
        sys.addaudithook(self.__sys_audit_hook)

    def __sys_audit_hook(self, event, *args):
        try:
            if self.__enabled and event == "open":
                filename, mode, flags = args[0]

                # Create a key for the file_open_counter that we will increment
                key = filename
                if self.track_mode:
                    key += f" (mode='{mode}')"
                if self.track_flags:
                    key += f" (flags='{hex(flags)}')"

                self.file_open_counter[key] += 1

        except Exception as ex:
            # In theory we shouldn't reach this point, but if we don't include
            # this try-catch block then we could hit an infinite loop if an
            # error *does* occur.
            warnings.warn(
                f"{ex.__class__.__name__} raised while calling {self.__class__.__name__}'s audit hook: {ex}"
            )

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> None:
        # Set __enabled so that we can enter the both of __sys_audit_hook
        self.__enabled = True

    def posthook(
        self,
        event_name: str,
        result: t.Any,
        context: None,
    ) -> None:
        self.__enabled = False

    def reset(self) -> None:
        self.file_open_counter.clear()

    def log_results(self, logger: logging.Logger) -> None:
        logger.info("%s results (file opened, count):", self.__class__.__name__)
        for (key, count) in self.file_open_counter.items():
            logger.info("    %s: %d", key, count)



import logging
import typing as t

# Type variable for contexts returned by prehooks
C = t.TypeVar("C")


class ProtoHook(t.Protocol[C]):
    """Interface for hooks that can be used by Seagrass."""

    # Priority in which prehooks and posthooks should be executed.
    #
    # - Prehooks with high priority are executed _after_
    #   hooks with a lower prehook_priority
    # - Posthooks with high priority are executed _before_
    #   hooks with lower posthook_priority
    #
    # This ensures that a ProtoHook whose prehook_priority and
    # posthook_priority are both high is likely to have its prehook
    # and posthook executed directly before and directly after the
    # execution of a wrapped event.
    prehook_priority: int = 0
    posthook_priority: int = 0

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> C:
        """Run the prehook."""
        ...

    def posthook(self, event_name: str, result: t.Any, context: C):
        """Run the posthook."""
        ...

    def reset(self) -> None:
        """Resets the internal state of the hook (if there is any)."""
        ...


@t.runtime_checkable
class LoggableHook(t.Protocol):
    """A mixin class for hooks that support an additional `log_results` method that
    outputs the results of the hook."""

    def log_results(
        self,
        logger: logging.Logger,
    ) -> None:
        """Log results that have been accumulated by the hook using the provided logger."""
        ...



import cProfile as prof
import logging
import pstats
import typing as t
from io import StringIO


class ProfilerHook:

    profiler: prof.Profile
    restrictions: t.Tuple[t.Any, ...]

    # Set a high prehook_priority and posthook_priority to ensure
    # that the profiler only gets called directly before and after
    # the event.
    prehook_priority: int = 10
    posthook_priority: int = 10

    def __init__(self, restrictions: t.Optional[t.Tuple[t.Any, ...]] = None):
        restrictions = tuple() if restrictions is None else restrictions
        self.restrictions = restrictions
        self.reset()

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> None:
        # Start profiling
        self.profiler.enable()

    def posthook(self, event_name: str, result: t.Any, context: None):
        # Stop profiling
        self.profiler.disable()

    def get_stats(self, **kwargs) -> pstats.Stats:
        """Return the profiling statistics as a pstats.Stats class."""
        return pstats.Stats(self.profiler, **kwargs)

    def reset(self) -> None:
        """Reset the internal profiler."""
        self.profiler = prof.Profile()

    def log_results(self, logger: logging.Logger):
        """Log the results captured by ProfilerHook."""
        # Dump results to an in-memory stream
        output = StringIO()
        stats = self.get_stats(stream=output)
        stats.print_stats(*self.restrictions)

        # Now take results from the in-memory stream and log them using the provided logger.
        logger.info("Results from %s:", self.__class__.__name__)
        logger.info("")
        output.seek(0)
        for line in output.readlines():
            logger.info("    " + line.rstrip())



# Tests for the LoggerHook auditing hook.

import logging
import unittest
from functools import reduce
from io import StringIO
from seagrass import Auditor
from seagrass.hooks import LoggingHook
from operator import add, mul


class LoggingHookTestCase(unittest.TestCase):
    def setUp(self):
        self.logging_output = StringIO()
        self.logger = logging.getLogger("logging_test_case")
        self.logger.setLevel(logging.DEBUG)

        fh = logging.StreamHandler(self.logging_output)
        fh.setLevel(logging.DEBUG)

        formatter = logging.Formatter("%(message)s")
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)

    def test_hook_function(self):
        auditor = Auditor(logger=self.logger)
        hook_pre = LoggingHook(
            prehook_msg=lambda e, args, kwargs: f"(prehook) hook_pre: {e}, {args=}, {kwargs=}"
        )
        hook_both = LoggingHook(
            prehook_msg=lambda e, args, kwargs: f"(prehook) hook_both: {e}, {args=}, {kwargs=}",
            posthook_msg=lambda e, result: f"(posthook) hook_both: {e}, {result=}",
        )

        event = "test.multiply_or_add"

        @auditor.decorate(event, hooks=[hook_pre, hook_both])
        def multiply_or_add(*args, op="*"):
            if op == "*":
                return reduce(mul, args, 1)
            elif op == "+":
                return reduce(add, args, 0)
            else:
                raise ValueError(f"Unknown operation '{op}'")

        args = (1, 2, 3, 4)
        kwargs_add = {"op": "+"}
        with auditor.audit():
            multiply_or_add(*args)
            multiply_or_add(*args, **kwargs_add)

        self.logging_output.seek(0)
        output = self.logging_output.read().rstrip().split("\n")
        self.assertEqual(
            output[0], f"(prehook) hook_pre: {event}, {args=}, kwargs={{}}"
        )
        self.assertEqual(
            output[1], f"(prehook) hook_both: {event}, {args=}, kwargs={{}}"
        )
        self.assertEqual(output[2], f"(posthook) hook_both: {event}, result={24}")
        self.assertEqual(
            output[3], f"(prehook) hook_pre: {event}, {args=}, kwargs={kwargs_add}"
        )
        self.assertEqual(
            output[4], f"(prehook) hook_both: {event}, {args=}, kwargs={kwargs_add}"
        )
        self.assertEqual(output[5], f"(posthook) hook_both: {event}, result={10}")



import unittest
from seagrass import Auditor
from seagrass.hooks import StackTraceHook


class StackTraceHookTestCase(unittest.TestCase):
    """Tests for the StackTraceHook auditing hook."""

    def test_hook_function(self):
        auditor = Auditor()
        hook = StackTraceHook()

        @auditor.decorate("test.foo", hooks=[hook])
        def foo():
            return

        @auditor.decorate("test.bar", hooks=[hook])
        def bar():
            return foo()

        with auditor.audit():
            foo()
            bar()

        # There are two unique stack traces for where test.foo gets
        # called, but only one for test.bar.
        self.assertEqual(len(hook.stack_trace_counter["test.foo"]), 2)
        self.assertEqual(len(hook.stack_trace_counter["test.bar"]), 1)

        hook.reset()
        self.assertEqual(len(hook.stack_trace_counter["test.foo"]), 0)
        self.assertEqual(len(hook.stack_trace_counter["test.bar"]), 0)


if __name__ == "__main__":
    unittest.main()



import inspect
import typing as t
from collections import Counter, defaultdict


class TracedFrame(t.NamedTuple):
    """A NamedTuple that stores frame information collected from stack
    traces. We have to use this in place of storing FrameInfo references
    since doing so can create reference cycles; see
    https://docs.python.org/3/library/inspect.html#the-interpreter-stack
    """

    filename: str
    lineno: int

    @classmethod
    def from_frame_info(cls, frame: inspect.FrameInfo) -> "TracedFrame":
        return TracedFrame(frame.filename, frame.lineno)

    def __str__(self):
        return f"{self.filename}#{self.lineno}"


class StackTraceHook:
    """An audit hook that captures the stack trace of where events are raised
    and collects statistics about caller locations."""

    prehook_priority: int = 0
    posthook_priority: int = 0

    stack_depth: t.Optional[int]
    stack_trace_counter: t.DefaultDict[str, t.Counter[t.Tuple[TracedFrame, ...]]]

    def __init__(self, stack_depth: t.Optional[int] = 10):
        self.stack_depth = stack_depth
        self.stack_trace_counter = defaultdict(Counter)

    def prehook(
        self, event: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> None:
        current_stack = inspect.stack()

        try:
            stack_depth = (
                len(current_stack) if self.stack_depth is None else self.stack_depth
            )
            frames = tuple(
                TracedFrame.from_frame_info(f) for f in current_stack[:stack_depth]
            )
            self.stack_trace_counter[event][frames] += 1
        finally:
            # See note in https://docs.python.org/3/library/inspect.html#the-interpreter-stack
            del current_stack

    def posthook(self, event: str, result: t.Any, context: None) -> None:
        # Posthook does nothing
        pass

    def reset(self):
        self.stack_trace_counter.clear()



# Tests for the CounterHook auditing hook.

from seagrass import Auditor
from seagrass.hooks import CounterHook
import unittest


class CounterHookTestCase(unittest.TestCase):
    def test_hook_function(self):
        auditor = Auditor()
        hook = CounterHook()

        @auditor.decorate("test.say_hello", hooks=[hook])
        def say_hello(name: str) -> str:
            return f"Hello, {name}!"

        self.assertEqual(hook.event_counter["test.say_hello"], 0)

        # Hook should not get called outside of an auditing context
        say_hello("Alice")
        self.assertEqual(hook.event_counter["test.say_hello"], 0)

        with auditor.audit():
            for name in ("Alice", "Bob", "Cathy"):
                say_hello(name)
        self.assertEqual(hook.event_counter["test.say_hello"], 3)
        self.assertEqual(set(hook.event_counter), set(("test.say_hello",)))

        # Upon resetting the hook, all event counts should be set back to zero
        hook.reset()
        self.assertEqual(hook.event_counter["test.say_hello"], 0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="seagrass",
    version="0.1.1",
    description="Auditing and profiling multi-tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="kernelmethod",
    author_email="17100608+kernelmethod@users.noreply.github.com",
    project_urls={
        "Source Code": "https://github.com/kernelmethod/Seagrass",
        "Bug Tracker": "https://github.com/kernelmethod/Seagrass/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    url="https://github.com/kernelmethod/Seagrass/",
    packages=find_packages(exclude=["tests"]),
    install_requires=[],
    python_requires=">=3.8.0",
    license="BSD",
)



# flake8: noqa: F401
from .counter_hook import CounterHook
from .file_open_hook import FileOpenHook
from .logging_hook import LoggingHook
from .stack_trace_hook import StackTraceHook
from .profiler_hook import ProfilerHook

__all__ = [
    "CounterHook",
    "FileOpenHook",
    "LoggingHook",
    "ProfilerHook",
    "StackTraceHook",
]



# Tests for raising Python audit events with sys.audit through Auditors

import sys
import unittest
import warnings
from collections import Counter, defaultdict
from seagrass import Auditor
from seagrass.hooks import CounterHook


class EventsTestCase(unittest.TestCase):
    """Tests for events created by Seagrass."""

    def setUp(self):
        self.auditor = Auditor()

    def test_wrap_class_property(self):
        # Override a class property to call a hook whenever it's accessed
        class Foo:
            def __init__(self):
                self.x = 0

            def add_one(self):
                return self.x + 1

        hook = CounterHook()

        @self.auditor.decorate("test.foo.get_x", hooks=[hook])
        def get_x(self):
            return self.__x

        @self.auditor.decorate("test.foo.set_x", hooks=[hook])
        def set_x(self, val):
            self.__x = val

        @self.auditor.decorate("test.foo.del_x", hooks=[hook])
        def del_x(self):
            del self.__x

        setattr(Foo, "x", property(fget=get_x, fset=set_x, fdel=del_x))

        with self.auditor.audit():
            f = Foo()
            f.x = 1
            y = f.x  # noqa: F841
            f.x += 2
            del f.x

        # We call get_x twice (once for y = f.x, another for f.x += 2)
        # We call set_x three times (once during Foo.__init__, once during f.x = 1, and
        #   once during f.x += 2)
        # We call del_x once (when we call del f.x)
        self.assertEqual(hook.event_counter["test.foo.get_x"], 2)
        self.assertEqual(hook.event_counter["test.foo.set_x"], 3)
        self.assertEqual(hook.event_counter["test.foo.del_x"], 1)

        # Now override the add_one function belonging to Foo
        current_add_one = Foo.add_one

        @self.auditor.decorate("test.foo.add_one", hooks=[hook])
        def add_one(self, *args, **kwargs):
            return current_add_one(self, *args, **kwargs)

        setattr(Foo, "add_one", add_one)

        with self.auditor.audit():
            f = Foo()
            result = f.add_one()

        self.assertEqual(result, 1)
        self.assertEqual(hook.event_counter["test.foo.add_one"], 1)

    def test_toggle_event(self):
        hook = CounterHook()

        @self.auditor.decorate("test.foo", hooks=[hook])
        def foo():
            return

        @self.auditor.decorate("test.bar", hooks=[hook])
        def bar():
            return foo()

        with self.auditor.audit():
            bar()
            self.assertEqual(hook.event_counter["test.foo"], 1)
            self.assertEqual(hook.event_counter["test.bar"], 1)

            # After disabling an event, its event hooks should no longer be called
            self.auditor.toggle_event("test.foo", False)
            bar()
            self.assertEqual(hook.event_counter["test.foo"], 1)
            self.assertEqual(hook.event_counter["test.bar"], 2)

            # Now we re-enable the event so that hooks get called again
            self.auditor.toggle_event("test.foo", True)
            bar()
            self.assertEqual(hook.event_counter["test.foo"], 2)
            self.assertEqual(hook.event_counter["test.bar"], 3)

    def test_wrap_function_and_create_sys_audit_event(self):
        # We should be able to set up sys.audit events when we wrap functions
        @self.auditor.decorate("test.foo", raise_audit_event=True)
        def foo(x, y, z=None):
            return x + y + (0 if z is None else z)

        @self.auditor.decorate("test.bar", raise_audit_event=False)
        def bar(x, y, z=None):
            return x + y + (0 if z is None else z)

        @self.auditor.decorate(
            "test.baz",
            raise_audit_event=True,
            prehook_audit_event_name="baz_prehook",
            posthook_audit_event_name="baz_posthook",
        )
        def baz(x, y, z=None):
            return x + y + (0 if z is None else z)

        events_counter = Counter()
        args_dict = defaultdict(list)

        def audit_hook(event: str, *args):
            try:
                if event.startswith("prehook:") or event.startswith("posthook:"):
                    events_counter[event] += 1
                    args_dict[event].append(args)
                elif event in ("baz_prehook", "baz_posthook"):
                    events_counter[event] += 1
                    args_dict[event].append(args)
            except Exception as ex:
                warnings.warn(f"Exception raised in audit_hook: {ex=}")

        sys.addaudithook(audit_hook)

        test_args = [(-3, 4), (5, 8), (0, 0)]
        test_kwargs = [{}, {}, {"z": 1}]

        def run_fns(args_list, kwargs_list):
            for (args, kwargs) in zip(args_list, kwargs_list):
                for fn in (foo, bar, baz):
                    fn(*args, **kwargs)

        # The following call to run_fns shouldn't raise any audit events since
        # it isn't performed in an auditing context.
        run_fns(test_args, test_kwargs)
        self.assertEqual(set(events_counter), set())
        self.assertEqual(set(args_dict), set())

        # Now some audit events should be raised:
        with self.auditor.audit():
            run_fns(test_args, test_kwargs)

        expected_prehooks = ["prehook:test.foo", "baz_prehook"]
        expected_posthooks = ["posthook:test.foo", "baz_posthook"]
        self.assertEqual(
            set(events_counter), set(expected_prehooks + expected_posthooks)
        )
        self.assertEqual(set(events_counter), set(args_dict))

        for event in expected_prehooks:
            self.assertEqual(events_counter[event], len(test_args))
            args = [args[0][0] for args in args_dict[event]]
            kwargs = [args[0][1] for args in args_dict[event]]
            self.assertEqual(args, test_args)
            self.assertEqual(kwargs, test_kwargs)

        # If we try running our functions outside of an auditing context again,
        # we should once again find that no system events are raised.
        events_counter.clear()
        args_dict.clear()
        run_fns(test_args, test_kwargs)
        self.assertEqual(set(events_counter), set())
        self.assertEqual(set(args_dict), set())


if __name__ == "__main__":
    unittest.main()



# Testing utilities and base classes for testing Seagrass

import logging
import unittest
from io import StringIO
from seagrass import Auditor


class SeagrassTestCaseBase(unittest.TestCase):
    def setUp(self):
        # Set up an auditor with a basic logging configuration
        self.logging_output = StringIO()
        fh = logging.StreamHandler(self.logging_output)
        fh.setLevel(logging.DEBUG)

        formatter = logging.Formatter("(%(levelname)s) %(message)s")
        fh.setFormatter(formatter)

        self.logger = logging.getLogger("test.seagrass")
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(fh)

        # Create a new auditor instance with the logger we just
        # set up
        self.auditor = Auditor(logger=self.logger)


class HookTestCaseBase(SeagrassTestCaseBase):
    """A base testing class for auditor hooks."""

    def setUp(self):
        super().setUp()

        if hasattr(self, "hook_gen"):
            self.hook = self.hook_gen()
        elif hasattr(self, "hook_class"):
            self.hook = self.hook_class()
        else:
            self.fail(
                "Either 'hook_gen' or 'hook_class' must be defined for children of BaseHookTestCase"
            )



# Tests for the FileOpenHook auditing hook.

import tempfile
import unittest
from test.base import HookTestCaseBase
from seagrass.hooks import FileOpenHook


class FileOpenHookTestCase(HookTestCaseBase):

    hook_class = FileOpenHook

    def test_hook_function(self):
        @self.auditor.decorate("test.say_hello", hooks=[self.hook])
        def say_hello(filename, name) -> str:
            with open(filename, "w") as f:
                f.write(f"Hello, {name}!\n")

            with open(filename, "r") as f:
                return f.read()

        with tempfile.NamedTemporaryFile() as f:
            with self.auditor.audit():
                result = say_hello(f.name, "Alice")

            self.assertEqual(result, "Hello, Alice!\n")
            self.assertEqual(self.hook.file_open_counter[f"{f.name} (mode='w')"], 1)
            self.assertEqual(self.hook.file_open_counter[f"{f.name} (mode='r')"], 1)

            # Check the logging output
            self.auditor.log_results()
            self.logging_output.seek(0)
            lines = [line.rstrip() for line in self.logging_output.readlines()]
            self.assertEqual(len(lines), 3)
            self.assertEqual(lines[1], f"(INFO)     {f.name} (mode='w'): 1")
            self.assertEqual(lines[2], f"(INFO)     {f.name} (mode='r'): 1")


if __name__ == "__main__":
    unittest.main()
